@@ -7,9 +7,13 @@ def get_current_request():
     return getattr(_thread_locals, 'request', None)
 
 
-class ThreadLocals(object):
+class ThreadLocals:
     """
     Middleware that stores the request object in thread local storage.
     """
-    def process_request(self, request):
+    def __init__(self, get_response):
+        self.get_response = get_response
+
+    def __call__(self, request):
         _thread_locals.request = request
+        return self.get_response(request)
