@@ -1,43 +1,48 @@
-from binascii import unhexlify
+import logging
+import warnings
 from base64 import b32encode
-
-from django.conf import settings
-from django.contrib.auth import login as login, REDIRECT_FIELD_NAME
-from django.contrib.auth.decorators import login_required
-from django.contrib.auth.forms import AuthenticationForm
-from django.contrib.sites.models import get_current_site
-from django.core.urlresolvers import reverse
-from django.forms import Form
-from django.http import HttpResponse, Http404
-from django.shortcuts import redirect
-from django.views.decorators.cache import never_cache
-from django.views.decorators.debug import sensitive_post_parameters
-from django.views.generic import FormView, DeleteView, TemplateView
-from django.views.generic.base import View
+from binascii import unhexlify
 
 import django_otp
+import qrcode
+import qrcode.image.svg
+from django.conf import settings
+from django.contrib.auth import REDIRECT_FIELD_NAME, login
+from django.contrib.auth.decorators import login_required
+from django.contrib.auth.forms import AuthenticationForm
+from django.contrib.sites.shortcuts import get_current_site
+from django.forms import Form
+from django.http import Http404, HttpResponse
+from django.shortcuts import redirect, resolve_url
+from django.urls import reverse
+from django.utils.http import is_safe_url
+from django.utils.module_loading import import_string
+from django.views.decorators.cache import never_cache
+from django.views.decorators.debug import sensitive_post_parameters
+from django.views.generic import DeleteView, FormView, TemplateView
+from django.views.generic.base import View
 from django_otp.decorators import otp_required
-from django_otp.plugins.otp_static.models import StaticToken, StaticDevice
-from django_otp.util import random_hex
+from django_otp.plugins.otp_static.models import StaticDevice, StaticToken
+
 from two_factor import signals
+from two_factor.models import get_available_methods, random_hex_str
 from two_factor.utils import totp_digits
 
+from ..forms import (
+    AuthenticationTokenForm, BackupTokenForm, DeviceValidationForm, MethodForm,
+    PhoneNumberForm, PhoneNumberMethodForm, TOTPDeviceForm, YubiKeyDeviceForm,
+)
+from ..models import PhoneDevice, get_available_phone_methods
+from ..utils import backup_phones, default_device, get_otpauth_url
+from .utils import IdempotentSessionWizardView, class_view_decorator
+
 try:
     from otp_yubikey.models import ValidationService, RemoteYubikeyDevice
 except ImportError:
     ValidationService = RemoteYubikeyDevice = None
 
-import qrcode
-import qrcode.image.svg
 
-from ..compat import is_safe_url, import_by_path
-from ..forms import (MethodForm, TOTPDeviceForm, PhoneNumberMethodForm,
-                     DeviceValidationForm, AuthenticationTokenForm,
-                     PhoneNumberForm, BackupTokenForm, YubiKeyDeviceForm)
-from ..models import PhoneDevice, get_available_phone_methods
-from ..utils import (get_otpauth_url, default_device,
-                     backup_phones)
-from .utils import (IdempotentSessionWizardView, class_view_decorator)
+logger = logging.getLogger(__name__)
 
 
 @class_view_decorator(sensitive_post_parameters())
@@ -78,7 +83,7 @@ class LoginView(IdempotentSessionWizardView):
     redirect_field_name = REDIRECT_FIELD_NAME
 
     def __init__(self, **kwargs):
-        super(LoginView, self).__init__(**kwargs)
+        super().__init__(**kwargs)
         self.user_cache = None
         self.device_cache = None
 
@@ -91,7 +96,7 @@ class LoginView(IdempotentSessionWizardView):
         if 'challenge_device' in self.request.POST:
             return self.render_goto_step('token')
 
-        return super(LoginView, self).post(*args, **kwargs)
+        return super().post(*args, **kwargs)
 
     def done(self, form_list, **kwargs):
         """
@@ -99,9 +104,13 @@ class LoginView(IdempotentSessionWizardView):
         """
         login(self.request, self.get_user())
 
-        redirect_to = self.request.GET.get(self.redirect_field_name, '')
-        if not is_safe_url(url=redirect_to, host=self.request.get_host()):
-            redirect_to = str(settings.LOGIN_REDIRECT_URL)
+        redirect_to = self.request.POST.get(
+            self.redirect_field_name,
+            self.request.GET.get(self.redirect_field_name, '')
+        )
+
+        if not is_safe_url(url=redirect_to, allowed_hosts=[self.request.get_host()]):
+            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)
 
         device = getattr(self.get_user(), 'otp_device', None)
         if device:
@@ -113,6 +122,10 @@ class LoginView(IdempotentSessionWizardView):
         """
         AuthenticationTokenForm requires the user kwarg.
         """
+        if step == 'auth':
+            return {
+                'request': self.request
+            }
         if step in ('token', 'backup'):
             return {
                 'user': self.get_user(),
@@ -147,7 +160,7 @@ class LoginView(IdempotentSessionWizardView):
         """
         if self.steps.current == 'token':
             self.get_device().generate_challenge()
-        return super(LoginView, self).render(form, **kwargs)
+        return super().render(form, **kwargs)
 
     def get_user(self):
         """
@@ -164,7 +177,7 @@ class LoginView(IdempotentSessionWizardView):
         """
         Adds user's default and backup OTP devices to the context.
         """
-        context = super(LoginView, self).get_context_data(form, **kwargs)
+        context = super().get_context_data(form, **kwargs)
         if self.steps.current == 'token':
             context['device'] = self.get_device()
             context['other_devices'] = [
@@ -176,7 +189,14 @@ class LoginView(IdempotentSessionWizardView):
             except StaticDevice.DoesNotExist:
                 context['backup_tokens'] = 0
 
-        context['cancel_url'] = settings.LOGOUT_URL
+        if getattr(settings, 'LOGOUT_REDIRECT_URL', None):
+            context['cancel_url'] = resolve_url(settings.LOGOUT_REDIRECT_URL)
+        elif getattr(settings, 'LOGOUT_URL', None):
+            warnings.warn(
+                "LOGOUT_URL has been replaced by LOGOUT_REDIRECT_URL, please "
+                "review the URL and update your settings.",
+                DeprecationWarning)
+            context['cancel_url'] = resolve_url(settings.LOGOUT_URL)
         return context
 
 
@@ -194,7 +214,7 @@ class SetupView(IdempotentSessionWizardView):
     is asked to provide a generated token. For call and sms methods, the user
     provides the phone number which is then validated in the final step.
     """
-    redirect_url = 'two_factor:setup_complete'
+    success_url = 'two_factor:setup_complete'
     qrcode_url = 'two_factor:qr'
     template_name = 'two_factor/core/setup.html'
     session_key_name = 'django_two_factor-qr_secret_key'
@@ -228,8 +248,20 @@ class SetupView(IdempotentSessionWizardView):
         Start the setup wizard. Redirect if already enabled.
         """
         if default_device(self.request.user):
-            return redirect(self.redirect_url)
-        return super(SetupView, self).get(request, *args, **kwargs)
+            return redirect(self.success_url)
+        return super().get(request, *args, **kwargs)
+
+    def get_form_list(self):
+        """
+        Check if there is only one method, then skip the MethodForm from form_list
+        """
+        form_list = super().get_form_list()
+        available_methods = get_available_methods()
+        if len(available_methods) == 1:
+            form_list.pop('method', None)
+            method_key, _ = available_methods[0]
+            self.storage.validated_step_data['method'] = {'method': method_key}
+        return form_list
 
     def render_next_step(self, form, **kwargs):
         """
@@ -237,13 +269,24 @@ class SetupView(IdempotentSessionWizardView):
         """
         next_step = self.steps.next
         if next_step == 'validation':
-            self.get_device().generate_challenge()
-        return super(SetupView, self).render_next_step(form, **kwargs)
+            try:
+                self.get_device().generate_challenge()
+                kwargs["challenge_succeeded"] = True
+            except Exception:
+                logger.exception("Could not generate challenge")
+                kwargs["challenge_succeeded"] = False
+        return super().render_next_step(form, **kwargs)
 
     def done(self, form_list, **kwargs):
         """
         Finish the wizard. Save all forms and redirect.
         """
+        # Remove secret key used for QR code generation
+        try:
+            del self.request.session[self.session_key_name]
+        except KeyError:
+            pass
+
         # TOTPDeviceForm
         if self.get_method() == 'generator':
             form = [form for form in form_list if isinstance(form, TOTPDeviceForm)][0]
@@ -258,7 +301,7 @@ class SetupView(IdempotentSessionWizardView):
             raise NotImplementedError("Unknown method '%s'" % self.get_method())
 
         django_otp.login(self.request, device)
-        return redirect(self.redirect_url)
+        return redirect(self.success_url)
 
     def get_form_kwargs(self, step=None):
         kwargs = {}
@@ -310,12 +353,12 @@ class SetupView(IdempotentSessionWizardView):
         self.storage.extra_data.setdefault('keys', {})
         if step in self.storage.extra_data['keys']:
             return self.storage.extra_data['keys'].get(step)
-        key = random_hex(20).decode('ascii')
+        key = random_hex_str(20)
         self.storage.extra_data['keys'][step] = key
         return key
 
     def get_context_data(self, form, **kwargs):
-        context = super(SetupView, self).get_context_data(form, **kwargs)
+        context = super().get_context_data(form, **kwargs)
         if self.steps.current == 'generator':
             key = self.get_key('generator')
             rawkey = unhexlify(key.encode('ascii'))
@@ -326,14 +369,14 @@ class SetupView(IdempotentSessionWizardView):
             })
         elif self.steps.current == 'validation':
             context['device'] = self.get_device()
-        context['cancel_url'] = settings.LOGIN_REDIRECT_URL
+        context['cancel_url'] = resolve_url(settings.LOGIN_REDIRECT_URL)
         return context
 
     def process_step(self, form):
         if hasattr(form, 'metadata'):
             self.storage.extra_data.setdefault('forms', {})
             self.storage.extra_data['forms'][self.steps.current] = form.metadata
-        return super(SetupView, self).process_step(form)
+        return super().process_step(form)
 
     def get_form_metadata(self, step):
         self.storage.extra_data.setdefault('forms', {})
@@ -352,7 +395,7 @@ class BackupTokensView(FormView):
     a pillow ;-).
     """
     form_class = Form
-    redirect_url = 'two_factor:backup_tokens'
+    success_url = 'two_factor:backup_tokens'
     template_name = 'two_factor/core/backup_tokens.html'
     number_of_tokens = 10
 
@@ -360,7 +403,7 @@ class BackupTokensView(FormView):
         return self.request.user.staticdevice_set.get_or_create(name='backup')[0]
 
     def get_context_data(self, **kwargs):
-        context = super(BackupTokensView, self).get_context_data(**kwargs)
+        context = super().get_context_data(**kwargs)
         context['device'] = self.get_device()
         return context
 
@@ -373,7 +416,7 @@ class BackupTokensView(FormView):
         for n in range(self.number_of_tokens):
             device.token_set.create(token=StaticToken.random_token())
 
-        return redirect(self.redirect_url)
+        return redirect(self.success_url)
 
 
 @class_view_decorator(never_cache)
@@ -388,19 +431,27 @@ class PhoneSetupView(IdempotentSessionWizardView):
     numbers can be used for verification.
     """
     template_name = 'two_factor/core/phone_register.html'
-    redirect_url = None
+    success_url = settings.LOGIN_REDIRECT_URL
     form_list = (
         ('setup', PhoneNumberMethodForm),
         ('validation', DeviceValidationForm),
     )
     key_name = 'key'
 
+    def get(self, request, *args, **kwargs):
+        """
+        Start the setup wizard. Redirect if no phone methods available.
+        """
+        if not get_available_phone_methods():
+            return redirect(self.success_url)
+        return super().get(request, *args, **kwargs)
+
     def done(self, form_list, **kwargs):
         """
         Store the device and redirect to profile page.
         """
         self.get_device(user=self.request.user, name='backup').save()
-        return redirect(self.redirect_url or str(settings.LOGIN_REDIRECT_URL))
+        return redirect(self.success_url)
 
     def render_next_step(self, form, **kwargs):
         """
@@ -409,7 +460,7 @@ class PhoneSetupView(IdempotentSessionWizardView):
         next_step = self.steps.next
         if next_step == 'validation':
             self.get_device().generate_challenge()
-        return super(PhoneSetupView, self).render_next_step(form, **kwargs)
+        return super().render_next_step(form, **kwargs)
 
     def get_form_kwargs(self, step=None):
         """
@@ -432,13 +483,13 @@ class PhoneSetupView(IdempotentSessionWizardView):
         The key is preserved between steps and stored as ascii in the session.
         """
         if self.key_name not in self.storage.extra_data:
-            key = random_hex(20).decode('ascii')
+            key = random_hex_str(20)
             self.storage.extra_data[self.key_name] = key
         return self.storage.extra_data[self.key_name]
 
     def get_context_data(self, form, **kwargs):
-        kwargs.setdefault('cancel_url', settings.LOGIN_REDIRECT_URL)
-        return super(PhoneSetupView, self).get_context_data(form, **kwargs)
+        kwargs.setdefault('cancel_url', resolve_url(self.success_url))
+        return super().get_context_data(form, **kwargs)
 
 
 @class_view_decorator(never_cache)
@@ -447,11 +498,13 @@ class PhoneDeleteView(DeleteView):
     """
     View for removing a phone number used for verification.
     """
+    success_url = settings.LOGIN_REDIRECT_URL
+
     def get_queryset(self):
         return self.request.user.phonedevice_set.filter(name='backup')
 
     def get_success_url(self):
-        return str(settings.LOGIN_REDIRECT_URL)
+        return resolve_url(self.success_url)
 
 
 @class_view_decorator(never_cache)
@@ -484,17 +537,19 @@ class QRGeneratorView(View):
         'SVG': 'image/svg+xml; charset=utf-8',
     }
 
+    def get_issuer(self):
+        return get_current_site(self.request).name
+
     def get(self, request, *args, **kwargs):
         # Get the data from the session
         try:
             key = self.request.session[self.session_key_name]
-            del self.request.session[self.session_key_name]
         except KeyError:
             raise Http404()
 
         # Get data for qrcode
         image_factory_string = getattr(settings, 'TWO_FACTOR_QR_FACTORY', self.default_qr_factory)
-        image_factory = import_by_path(image_factory_string)
+        image_factory = import_string(image_factory_string)
         content_type = self.image_content_types[image_factory.kind]
         try:
             username = self.request.user.get_username()
@@ -502,7 +557,7 @@ class QRGeneratorView(View):
             username = self.request.user.username
 
         otpauth_url = get_otpauth_url(accountname=username,
-                                      issuer=get_current_site(self.request).name,
+                                      issuer=self.get_issuer(),
                                       secret=key,
                                       digits=totp_digits())
 
