@@ -1,9 +1,11 @@
 import logging
-from django.core.exceptions import ValidationError
+
+from django.core.exceptions import SuspiciousOperation
 from django.utils.decorators import method_decorator
-from django.utils.functional import lazy_property
-from django.utils.translation import ugettext as _
-from two_factor.compat import ManagementForm, SessionStorage, SessionWizardView
+from django.utils.translation import gettext as _
+from formtools.wizard.forms import ManagementForm
+from formtools.wizard.storage.session import SessionStorage
+from formtools.wizard.views import SessionWizardView
 
 logger = logging.getLogger(__name__)
 
@@ -16,17 +18,23 @@ class ExtraSessionStorage(SessionStorage):
     validated_step_data_key = 'validated_step_data'
 
     def init_data(self):
-        super(ExtraSessionStorage, self).init_data()
+        super().init_data()
         self.data[self.validated_step_data_key] = {}
 
+    def reset(self):
+        if self.prefix in self.request.session:
+            super().reset()
+        else:
+            self.init_data()
+
     def _get_validated_step_data(self):
         return self.data[self.validated_step_data_key]
 
     def _set_validated_step_data(self, validated_step_data):
         self.data[self.validated_step_data_key] = validated_step_data
 
-    validated_step_data = lazy_property(_get_validated_step_data,
-                                        _set_validated_step_data)
+    validated_step_data = property(_get_validated_step_data,
+                                   _set_validated_step_data)
 
 
 class IdempotentSessionWizardView(SessionWizardView):
@@ -100,14 +108,11 @@ class IdempotentSessionWizardView(SessionWizardView):
         # Check if form was refreshed
         management_form = ManagementForm(self.request.POST, prefix=self.prefix)
         if not management_form.is_valid():
-            raise ValidationError(
-                _('ManagementForm data is missing or has been tampered.'),
-                code='missing_management_form',
-            )
+            raise SuspiciousOperation(_('ManagementForm data is missing or has been tampered with'))
 
         form_current_step = management_form.cleaned_data['current_step']
-        if (form_current_step != self.steps.current and
-                self.storage.current_step is not None):
+        if (form_current_step != self.steps.current
+                and self.storage.current_step is not None):
             # form refreshed, change current step
             self.storage.current_step = form_current_step
         # -- End duplicated code from upstream
@@ -120,7 +125,7 @@ class IdempotentSessionWizardView(SessionWizardView):
                            self.steps.current)
             return self.render_goto_step(self.steps.all[-1])
 
-        return super(IdempotentSessionWizardView, self).post(*args, **kwargs)
+        return super().post(*args, **kwargs)
 
     def process_step(self, form):
         """
@@ -146,7 +151,7 @@ class IdempotentSessionWizardView(SessionWizardView):
         for next_step in keys[key:]:
             self.storage.validated_step_data.pop(next_step, None)
 
-        return super(IdempotentSessionWizardView, self).process_step(form)
+        return super().process_step(form)
 
     def render_done(self, form, **kwargs):
         """
